#include "k8055.h"
#include <math.h>
#define STR_BUFF 256

/* set debug to 0 to not print excess info */
int DEBUG=0;

/* variables for usb */
struct usb_bus *bus, *busses;
struct usb_device *dev;
usb_dev_handle *device_handle = '\0';

/*globals for datatransfer */
unsigned char data_in[9],data_out[9];
/*char* device_id[]; */

static int takeover_device( usb_dev_handle* udev, int interface ){
  char driver_name[STR_BUFF];
  memset( driver_name, 0, STR_BUFF );
#if !defined _BSD
  int ret = -1;
#endif
  assert( udev != NULL );
#if !defined _BSD
  ret = usb_get_driver_np( udev, interface, driver_name, sizeof(driver_name) );
  if ( ret ==0 ){
    if (DEBUG) fprintf(stderr,"Got driver name: %s\n",driver_name);
      if ( 0 > usb_detach_kernel_driver_np( udev, interface ) )
      {
        if (DEBUG) fprintf( stderr, "Disconnect OS driver: %s\n", usb_strerror());
      }
      else
        if (DEBUG) fprintf( stderr, "Disconnected OS driver: %s\n", usb_strerror());
  }
  else
    if (DEBUG) fprintf( stderr, "get driver name: %s\n", usb_strerror());
#endif
  /*claim interface */
  if( usb_claim_interface(udev, interface) <0 )
  {
    if (DEBUG) fprintf( stderr,"Claim interface error: %s\n", usb_strerror());
    return -1;
  }
  else 
  usb_set_altinterface(udev, interface);
  usb_set_configuration(udev, 1);

  if (DEBUG) fprintf( stderr,"Found interface %d\n", interface);
  if (DEBUG) fprintf( stderr, "Took over the device\n");
  
  return 0;
}

int OpenDevice(long board_address){

  unsigned char located = 0;
  int ipid, open_status;

  /* init USB and find all of the devices on all busses */
  usb_init();
  usb_find_busses();
  usb_find_devices();
  busses = usb_get_busses();

  /*ID of the welleman board is 5500h + address config */
  if (board_address>=0 && board_address<4)
    ipid = 0x5500+(int)board_address;
  else
    return -1; /*throw error instead of being nice */

  /*start looping through the devices to find the correct one */
  for (bus = busses; bus; bus = bus->next) {
    for (dev = bus->devices; dev; dev = dev->next) {
      if ( (dev->descriptor.idVendor == 0x10cf) &&
           (dev->descriptor.idProduct == ipid ) ){
      	located++;
        device_handle = usb_open(dev);
        if (DEBUG) fprintf(stderr,"Velleman Device Found @ Address %s Vendor 0x0%x Product ID 0x0%x\n", 
		dev->filename, dev->descriptor.idVendor, dev->descriptor.idProduct);
        if (takeover_device( device_handle, 0 )<0 ){
          if (DEBUG)fprintf(stderr,"Can not take over the device from the OS driver\n");
	  usb_close(device_handle); /*close usb if we fail */
          return -1; /*throw -1 to show that OpenDevice failed */
	}
        else 
	  open_status=usb_interrupt_read(device_handle,0x81,(char *)data_in,8,20);
	if (open_status!=8)
	  return -1;
	else
  	  return data_in[1]-1;
      }
    }
  }
  if (DEBUG) fprintf(stderr,"Could not find velleman k8055 with address %d\n",(int)board_address);
  return -1;
}

int CloseDevice() {
  return usb_close(device_handle);
}

long ReadAnalogChannel(long Channelno){
	int open_status=0;
	open_status=usb_interrupt_read(device_handle,0x81,(char *)data_in,8,20);
	if (open_status==8){
  	  if (Channelno==2)
	    return data_in[3];
	  else
	    return data_in[2];
  	}
	else
	  return -1;
}

int ReadAllAnalog(long* data1, long* data2){
	int open_status=0;
	open_status=usb_interrupt_read(device_handle,0x81,(char *)data_in,8,20);
	if (open_status==8){
	  *data1=data_in[2];
	  *data2=data_in[3];
	  return 0;
	}
        else
          return -1;
}

int OutputAnalogChannel(long channel, long data){
	int open_status=0;
	data_out[0]=0x05; /* analog out or digital out command */
	if (channel==2)
	  data_out[3]=(unsigned char)data;
	else
	  data_out[2]=(unsigned char)data;
	open_status=usb_interrupt_write(device_handle,0x01,(char *)data_out,8,20);
	open_status=usb_interrupt_write(device_handle,0x01,(char *)data_out,8,20);
	if (open_status!=8)
	  return -1;
	else
	  return 0;
}

int OutputAllAnalog(long data1,long data2){
	int open_status=0;
	data_out[0]=0x05; /* analog out or digital out command */
	  data_out[2]=(unsigned char)data1;
	  data_out[3]=(unsigned char)data2;
	open_status=usb_interrupt_write(device_handle,0x01,(char *)data_out,8,20);
	open_status=usb_interrupt_write(device_handle,0x01,(char *)data_out,8,20);
	if (open_status!=8)
	  return -1;
	else
	  return 0;
}

int ClearAllAnalog(){
	return OutputAllAnalog(0,0);
}

int ClearAnalogChannel(long channel){
	if (channel==2)
	  return OutputAnalogChannel(2,0);
	else
	  return OutputAnalogChannel(1,0);
}

int SetAnalogChannel(long channel){
	if (channel==2)
          return OutputAnalogChannel(2,0xff);
        else
          return OutputAnalogChannel(1,0xff);

}

int SetAllAnalog(){
	return OutputAllAnalog(0xff,0xff);
}

int WriteAllDigital(long data){
        int open_status=0;
        data_out[0]=0x05; /* analog out or digital out command */
        data_out[1]=(unsigned char)data;
        open_status=usb_interrupt_write(device_handle,0x01,(char *)data_out,8,20);
        open_status=usb_interrupt_write(device_handle,0x01,(char *)data_out,8,20);
        if (open_status!=8)
          return -1;
        else
          return 0;
}

int ClearDigitalChannel(long channel){
	int i,mask=1;
	unsigned char data;
	if (channel>0 && channel <9){
	  for (i=1;i<channel;i++)
	    mask*=2;	
	  data=data_out[1] & (255-mask);
	  return WriteAllDigital(data);
	}
	else 
	  return -1;
}

int ClearAllDigital(){
	return WriteAllDigital(0x00);
}

int SetDigitalChannel(long channel){
	int i,mask=1;
	unsigned char data;
	if (channel>0 && channel <9){
	  for (i=1;i<channel;i++)
	    mask*=2;	
	  data=data_out[1] | mask;
	  return WriteAllDigital(data);
	}
	else 
	  return -1;
}

int SetAllDigital(){
	return WriteAllDigital(0xff);
}

int ReadDigitalChannel(long channel){
	int open_status=0,i;
	unsigned char mask=8;
	if (channel>0 && channel<6){
	  open_status=usb_interrupt_read(device_handle,0x81,(char *)data_in,8,20);
	  if (open_status==8){
	    for (i=1;i<=channel;i++)
	      if (i!=3) mask*=2;
	    if (channel==3)
	      mask=1;  /*for some reason channel 3 is 0x01, the others are
			 the msb like 0x10, 0x20 ... */	
	    return (data_in[0] & mask)>0?1:0;    
  	  }
	  else
	    return -1;
	}
	else
	  return -1;
}

long ReadAllDigital(){
	int open_status=0,i,return_data=0;
	unsigned char mask=8,pow=1;
	open_status=usb_interrupt_read(device_handle,0x81,(char *)data_in,8,20);
	if (open_status==8){
	    for (i=1;i<=5;i++){
	      mask*=2;
	      if (i==3){
	        return_data+=(4*(data_in[0] & 1));
		mask/=2;
	      }
	      else
	        if (data_in[0] & mask)
	          return_data+=pow;   
	      pow*=2; 
	    }
	  return return_data;
	}
	else
	  return -1;
}

int ResetCounter(long counternr){
        int open_status=0;
	
	if (counternr==1 || counternr==2){
          data_out[0]=0x02+(unsigned char)counternr; /* counter selection */
	  data_out[3+counternr]=0x00;
	  open_status=usb_interrupt_write(device_handle,0x01,(char *)data_out,8,20);
	  open_status=usb_interrupt_write(device_handle,0x01,(char *)data_out,8,20);
	  if (open_status==8)
	    return 0;
	  else
	    return -1;
	}
	else
	  return -1;
}

long ReadCounter(long counterno){
	int open_status=0;
	if (counterno==1 || counterno==2){
  	  open_status=usb_interrupt_read(device_handle,0x81,(char *)data_in,8,20);
	  if (open_status==8){
  	    if (counterno==2)
	      return data_in[6];
	    else
	      return data_in[4];
	  }
	  else
	    return -1;
  	}
	else
	  return -1;
}

int SetCounterDebounceTime(long counterno, long debouncetime){
	int open_status=0;	
	float value;
	if (counterno==1 || counterno==2){
	  data_out[0]=(unsigned char)counterno;
	  /*the velleman k8055 use a exponetial formula to split up the
	    debouncetime 0-7450 over value 1-255. I've tested every value
	    and found that the formula dbt=0,338*value^1,8017 is closest to
	    vellemans dll. By testing and measuring times on the other hand
	    I found the formula dbt=0,115*x^2 quite near the actual values, a
	    little below at really low values and a little above at really 
	    high values. But the time set with this formula is within +-4% */
	  if (debouncetime>7450)
	    debouncetime=7450; 
	  value=sqrtf(debouncetime/0.115);
	  if (value> ((int)value+0.49999999)) /*simple round() function) */
	    value+=1;
	  data_out[5+counterno]=(unsigned char)value;
	  if (DEBUG) fprintf(stderr,"Debouncetime%d value for k8055:%d\n",(int)counterno,data_out[5+counterno]);
  	  open_status=usb_interrupt_write(device_handle,0x01,(char *)data_out,8,20);
	  if (open_status==8)
  	    return 0;
	  else
	    return -1;
	}
	else
	  return -1;
}
